#include "gp_cuda.h"
#include <math.h>
#include <assert.h>

// Macro to catch CUDA errors in CUDA runtime calls
// Thanks to http://forums.nvidia.com/index.php?showtopic=183497&pid=1134164&mode=threaded&start=#entry1134164
#define CUDA_SAFE_CALL(call) \
do { \
cudaError_t err = call; \
if (cudaSuccess != err) { \
fprintf (stderr, "Cuda error in file '%s' in line %i : %s.\n",\
__FILE__, __LINE__, cudaGetErrorString(err) ); \
exit(EXIT_FAILURE); \
} \
} while (0)



//Floating point indicator pointer for GPU device
__constant__ float *fIndValsP;
//Boolean indicator pointer for GPU device
__constant__ char *bIndValsP;
//Space to store GP individuals on GPU
__constant__ char *individuals;
//Space to store start of individuals on GPU
__constant__ unsigned long int * dStartInds;

//Space to store money of individuals/stocks
float * moneyP;
//Space to store signals (not used anymore)
char * signalsP;
//Space to store ROI of individuals
float * roiP;

//Stack array for all individuals
float *shared_array;

//Calculates the ROI of each individual
//float * roi: Pointer to array that stores ROIs
//float * money: Pointer to array that stores money for each individual/stock
//unsigned long int popSize: The number of individuals
__global__ void calcROI(float * roi, float * money, unsigned long int popSize){
	//Get the number of the individual (one thread per individual)
	unsigned int indNum = ((blockIdx.x * blockDim.x) + (int)threadIdx.x);
	int i;
	//Return if not a valid individual
	if(indNum >= popSize){
		return;
	}

	//Initialize total profit to 0
	float totProf = 0;
	//For each stock, increase profit by profit made on that stockk
	for(i = 0; i < NUM_STOCKS; i++){
		totProf += money[(indNum * NUM_STOCKS) + i] - INIT_MONEY;
	}

	//Total ROI = Total Profit / Total Initial Money
	roi[indNum] = totProf / (INIT_MONEY * NUM_STOCKS);
}


//Stack declaration
#define STACK_SIZE 20
#define initStack float *Stack = shared_array;

//Macro for accessing a value on the stack (called by an individual on their own stack)
//Indexed by stack pointer then by thread number so conscutive threads will be accessing consecutive memory locations
#define stack(sp) Stack[((sp) * POP_SIZE * NUM_STOCKS) + threadNum]
//Pushes value x onto the individual's stack
#define push(x) {stack(SP) = x; SP++;}

//Evaluates the individuals over all stocks/days, updating the money variable so ROI can be calculated afterwards
//char * rpn: The RPN strings of all individuals
//float * money: The array of money values for each individual/stock combination
//unsigned long int maxInds: The number of individuals this generation
//float * fIndValsP: Pointer to the floating point indicators
//char * bIndValsP: Pointer to the boolean indicators
//float * sharedArray: The stack array for all threads
//unsigned long int * dStartInds: The starting locations in rpn of each individual
__global__ void generateSignals(char *rpn, float *money, unsigned long int maxInds, float * fIndValsP, char *bIndValsP, float *shared_array, unsigned long int* dStartInds){
	//initialize the stack
	initStack
	
	//Generate individual/stock/thread numbers based on indexes
	unsigned int indNum = (unsigned int)((blockIdx.x * blockDim.x) + (unsigned int)threadIdx.x)/NUM_STOCKS;
	unsigned int stockNum = (unsigned int)((blockIdx.x * blockDim.x) + (unsigned int)threadIdx.x)%NUM_STOCKS;
	unsigned int threadNum = ((blockIdx.x * blockDim.x) + threadIdx.x);
	unsigned long int moneyIndex = (indNum * NUM_STOCKS) + stockNum;
	
	//The signal of the current day (will vary: -1, 0, 1)
	char signal;
	//Current number of shares of the current stock owned by the current individual
	unsigned int numShares = 0;
	
	//Return if invalid thread
	if(indNum >= maxInds || stockNum >= NUM_STOCKS || threadNum >= (NUM_STOCKS * maxInds)){
		return;
	}

	//The current tree number (either 0/buy or 1/sell)
	int treeNum;
	//Maintains current spot in individuals rpn string
	int i;
	//The current day
	int day;
	
	//Initialize the money for the individual/stock to INIT_MONEY
	money[moneyIndex] = INIT_MONEY;
	
	//initialize the stack pointer and declare variables needed
	int SP = 0;
	char op;
	float SP1;
	float SP2;
	
	//For every day in the set of training days
	for(day = 0; day < TRAINING_DAYS; day++){
		//initialize the signal to 0 (no action)
		signal = 0;
		//For both the buy and sell trees
		for(treeNum = 0; treeNum < NUM_TREES; treeNum++){
			//stack pointer and current index in string are 0
			SP = 0;
			i = 0;
			//until we get to null string (end of invidiaul)
			while(true){
				//get the current operation code from the string
				op = rpn[i + dStartInds[(indNum * NUM_TREES) + (treeNum)]];

				//if null string, break
				if(op == 0){
					break;
				}
				
				//find which operation should be completed
				if(op == AND[0]){
					SP1 = stack(SP-1);
					SP2 = stack(SP-2);
					SP -= 2;
					push(SP1 && SP2);
				}else if(op == OR[0]){
					SP1 = stack(SP-1);
					SP2 = stack(SP-2);
					SP -= 2;
					push(SP1 || SP2);
				}else if(op == NOT[0]){
					SP1 = stack(SP-1);
					SP -= 1;
					push(!SP1);
				}else if(op == GT[0]){
					SP1 = stack(SP-2);
					SP2 = stack(SP-1);
					SP -= 2;
					push(SP1 > SP2);
				}else if(op == LT[0]){
					SP1 = stack(SP-2);
					SP2 = stack(SP-1);
					SP -= 2;
					push(SP1 < SP2);
				}else if(op == MA1[0]){
					push(fIndValsP[(day * NUM_FLOAT_INDS * NUM_STOCKS) + (vMA1 * NUM_STOCKS) + stockNum]);
				}else if(op == MA5[0]){
					push(fIndValsP[(day * NUM_FLOAT_INDS * NUM_STOCKS) + (vMA5 * NUM_STOCKS) + stockNum]);
				}else if(op == MA10[0]){
					push(fIndValsP[(day * NUM_FLOAT_INDS * NUM_STOCKS) + (vMA10 * NUM_STOCKS) + stockNum]);
				}else if(op == MA15[0]){
					push(fIndValsP[(day * NUM_FLOAT_INDS * NUM_STOCKS) + (vMA15 * NUM_STOCKS) + stockNum]);
				}else if(op == MA25[0]){
					push(fIndValsP[(day * NUM_FLOAT_INDS * NUM_STOCKS) + (vMA25 * NUM_STOCKS) + stockNum]);
				}else if(op == MA50[0]){
					push(fIndValsP[(day * NUM_FLOAT_INDS * NUM_STOCKS) + (vMA50 * NUM_STOCKS) + stockNum]);
				}else if(op == MA75[0]){
					push(fIndValsP[(day * NUM_FLOAT_INDS * NUM_STOCKS) + (vMA75 * NUM_STOCKS) + stockNum]);
				}else if(op == MA100[0]){
					push(fIndValsP[(day * NUM_FLOAT_INDS * NUM_STOCKS) + (vMA100 * NUM_STOCKS) + stockNum]);
				}else if(op == MA150[0]){
					push(fIndValsP[(day * NUM_FLOAT_INDS * NUM_STOCKS) + (vMA150 * NUM_STOCKS) + stockNum]);
				}else if(op == MA200[0]){
					push(fIndValsP[(day * NUM_FLOAT_INDS * NUM_STOCKS) + (vMA200 * NUM_STOCKS) + stockNum]);
				}else if(op == EA5[0]){
					push(fIndValsP[(day * NUM_FLOAT_INDS * NUM_STOCKS) + (vEMA5 * NUM_STOCKS) + stockNum]);
				}else if(op == EA9[0]){
					push(fIndValsP[(day * NUM_FLOAT_INDS * NUM_STOCKS) + (vEMA9 * NUM_STOCKS) + stockNum]);
				}else if(op == EA15[0]){
					push(fIndValsP[(day * NUM_FLOAT_INDS * NUM_STOCKS) + (vEMA15 * NUM_STOCKS) + stockNum]);
				}else if(op == EA20[0]){
					push(fIndValsP[(day * NUM_FLOAT_INDS * NUM_STOCKS) + (vEMA20 * NUM_STOCKS) + stockNum]);
				}else if(op == EA25[0]){
					push(fIndValsP[(day * NUM_FLOAT_INDS * NUM_STOCKS) + (vEMA25 * NUM_STOCKS) + stockNum]);
				}else if(op == CLOSING[0]){
					push(fIndValsP[(day * NUM_FLOAT_INDS * NUM_STOCKS) + (vCP * NUM_STOCKS) + stockNum]);
				}else if(op == TYPICAL[0]){
					push(fIndValsP[(day * NUM_FLOAT_INDS * NUM_STOCKS) + (vTP * NUM_STOCKS) + stockNum]);
				}else if(op == NVIG[0]){
					push(bIndValsP[(day * NUM_BOOL_INDS * NUM_STOCKS) + (vNVIG * NUM_STOCKS) + stockNum]);
				}else if(op == NVIL[0]){
					push(bIndValsP[(day * NUM_BOOL_INDS * NUM_STOCKS) + (vNVIL * NUM_STOCKS) + stockNum]);
				}else if(op == PVIG[0]){
					push(bIndValsP[(day * NUM_BOOL_INDS * NUM_STOCKS) + (vPVIG * NUM_STOCKS) + stockNum]);
				}else if(op == PVIL[0]){
					push(bIndValsP[(day * NUM_BOOL_INDS * NUM_STOCKS) + (vPVIL * NUM_STOCKS) + stockNum]);
				}else if(op == MACDL[0]){
					push(bIndValsP[(day * NUM_BOOL_INDS * NUM_STOCKS) + (vMACDL * NUM_STOCKS) + stockNum]);
				}else if(op == MACDG[0]){
					push(bIndValsP[(day * NUM_BOOL_INDS * NUM_STOCKS) + (vMACDG * NUM_STOCKS) + stockNum]);
				}else if(op == MACDGZ[0]){
					push(bIndValsP[(day * NUM_BOOL_INDS * NUM_STOCKS) + (vMACDGZ * NUM_STOCKS) + stockNum]);
				}else if(op == MACDLZ[0]){
					push(bIndValsP[(day * NUM_BOOL_INDS * NUM_STOCKS) + (vMACDLZ * NUM_STOCKS) + stockNum]);
				}else if(op == CCIG[0]){
					push(bIndValsP[(day * NUM_BOOL_INDS * NUM_STOCKS) + (vCCIG * NUM_STOCKS) + stockNum]);
				}else if(op == CCIL[0]){
					push(bIndValsP[(day * NUM_BOOL_INDS * NUM_STOCKS) + (vCCIL * NUM_STOCKS) + stockNum]);
				}else if(op == EOML[0]){
					push(bIndValsP[(day * NUM_BOOL_INDS * NUM_STOCKS) + (vEOML * NUM_STOCKS) + stockNum]);
				}else if(op == EOMG[0]){
					push(bIndValsP[(day * NUM_BOOL_INDS * NUM_STOCKS) + (vEOMG * NUM_STOCKS) + stockNum]);
				}else if(op == MFIG[0]){
					push(bIndValsP[(day * NUM_BOOL_INDS * NUM_STOCKS) + (vMFIG * NUM_STOCKS) + stockNum]);
				}else if(op == MFIL[0]){
					push(bIndValsP[(day * NUM_BOOL_INDS * NUM_STOCKS) + (vMFIL * NUM_STOCKS) + stockNum]);
				}else{
					//if an unexpected operation, break (this should never happen)
					break;
				}
				//go to next element in the individual rpn string
				i++;
			}

			//update the signal, this adds the value of the buy tree (0/1) and subtracts the value of the sell tree (0/1)
			signal += (char)((((-treeNum)*2)+1)*(char)stack(0));
		}
		
		//If the signal is > 0 and no shares owned, buy as many as possible
		if (signal > 0 && numShares <= 0){
			//First remove the commission fee
			money[moneyIndex] = money[moneyIndex] - COMMISION_FEE;
			//Calculate the number of shares we can afford
			numShares = floor(money[moneyIndex] / fIndValsP[(day * NUM_FLOAT_INDS * NUM_STOCKS) + (vCP * NUM_STOCKS) + stockNum]);
			//Update money
			money[moneyIndex] = money[moneyIndex] - (numShares * fIndValsP[(day * NUM_FLOAT_INDS * NUM_STOCKS) + (vCP * NUM_STOCKS) + stockNum]);
		//If the signal is < 0, sell all shares owned (if any)
		}else if (signal < 0 && numShares > 0){
			//Update the money
			money[moneyIndex] = money[moneyIndex] + (numShares * fIndValsP[(day * NUM_FLOAT_INDS * NUM_STOCKS) + (vCP * NUM_STOCKS) + stockNum]);
			//Remove commission fee
			money[moneyIndex] = money[moneyIndex] - COMMISION_FEE;
			//Set shares back to 0
			numShares = 0;
		}
	}
	
	//If we have shares at the end of training, sell them to get the current value of the portfolio
	if (numShares > 0){
		money[moneyIndex] = money[moneyIndex] + (numShares * fIndValsP[((TRAINING_DAYS - 1) * NUM_FLOAT_INDS * NUM_STOCKS) + (vCP * NUM_STOCKS) + stockNum]);
		numShares = 0;
	}
}

//Method called from app.c to evaluate a population on the gpu
//char * inds: The string of RPN individuals
//unsigned long int indsSize: The length of *inds
//unsigned long int numTrees: Number of individuals * number of trees(2)
//char * signals: Pointer in host memory for signals, was used to compare to sequential signals. Could be removed now but didnt want to go through all the code and take it out
//unsigned long int popSize: Number of individuals in the population to be evaluated
//float * roi: Pointer in host memory to copy ROI to for comparison to sequential evaluation
//unsigned long int * startInds: Array of start indexes in *inds for all individuals/trees
//int compareToSeq: Not needed anymore, like *signals
void eval_pop_gpu(char* inds, unsigned long int indsSize, unsigned long int numTrees, char *signals, unsigned long int popSize, float*roi,  unsigned long int * startInds, int compareToSeq){
	//Allocate memory for individuals/indexes and copy them to device
	CUDA_SAFE_CALL(cudaMalloc(&individuals, (indsSize) * sizeof(char)));
	CUDA_SAFE_CALL(cudaMemcpy(individuals, inds, indsSize * sizeof(char), cudaMemcpyHostToDevice));
	CUDA_SAFE_CALL(cudaMalloc(&dStartInds, numTrees * sizeof(unsigned long int)));
	CUDA_SAFE_CALL(cudaMemcpy(dStartInds, startInds, numTrees * sizeof(unsigned long int), cudaMemcpyHostToDevice));
	
	//Setup block size and number of blocks for evaluationg
	int blockSize = 512;
	float numThreads = NUM_STOCKS * popSize;
	int numBlocks = ceil((float)numThreads / blockSize);
	
	//Call the kernel that evaluates all individuals
	generateSignals<<< numBlocks, blockSize >>> (individuals, moneyP, popSize, fIndValsP, bIndValsP, shared_array, dStartInds);
	
	//Sync all threads, needed with Fermi because kernel calls return immediately
	CUDA_SAFE_CALL(cudaThreadSynchronize());
	
	//Free the individuals/indexes
	CUDA_SAFE_CALL(cudaFree(individuals));
	CUDA_SAFE_CALL(cudaFree(dStartInds));
	
	//Redefine number of threads and blocks
	numThreads = popSize;
	numBlocks = ceil((float)numThreads / blockSize);
	
	//Call kernel to calulate ROI of individuals
	calcROI<<< numBlocks, blockSize >>>(roiP, moneyP, popSize);
	cudaThreadSynchronize();
	
	//Copy ROI back to host memory so fitness can be calculated
	CUDA_SAFE_CALL(cudaMemcpy(roi, roiP, (popSize * sizeof(float)), cudaMemcpyDeviceToHost));
	cudaThreadSynchronize();
}

//Frees all memory allocated by copy_init_data
void uninit_data(){
	CUDA_SAFE_CALL(cudaFree(fIndValsP));
	CUDA_SAFE_CALL(cudaFree(bIndValsP));
	CUDA_SAFE_CALL(cudaFree(shared_array));
	CUDA_SAFE_CALL(cudaFree(moneyP));
	CUDA_SAFE_CALL(cudaFree(roiP));
}

//Allocates and copies data from host to GPU
void copy_init_data(float *fInds, char *bInds){
	//Allocate and copy floating point indicators
	CUDA_SAFE_CALL(cudaMalloc(&fIndValsP, (NUM_STOCKS * NUM_FLOAT_INDS * TRAINING_DAYS * sizeof(float))));
	CUDA_SAFE_CALL(cudaMemcpy(fIndValsP, fInds, (NUM_STOCKS * NUM_FLOAT_INDS * TRAINING_DAYS * sizeof(float)), cudaMemcpyHostToDevice));
	//Allocate and copy boolean indicators
	CUDA_SAFE_CALL(cudaMalloc(&bIndValsP, (NUM_STOCKS * NUM_BOOL_INDS * TRAINING_DAYS * sizeof(char))));
	CUDA_SAFE_CALL(cudaMemcpy(bIndValsP, bInds, (NUM_STOCKS * NUM_BOOL_INDS * TRAINING_DAYS * sizeof(char)), cudaMemcpyHostToDevice));

	//Allocate stack, money, and ROI arrays
	//This way we don't have to copy/allocate each population evaluation
	CUDA_SAFE_CALL(cudaMalloc(&shared_array, POP_SIZE*NUM_STOCKS*STACK_SIZE*sizeof(float)));
	CUDA_SAFE_CALL(cudaMalloc(&moneyP, (POP_SIZE * NUM_STOCKS * sizeof(float))));
	CUDA_SAFE_CALL(cudaMalloc(&roiP, (POP_SIZE * sizeof(float))));
}




